<?php
/**
 * @package     Server Cache for Joomla! Supports Nginix / Varnish
 * @version     1.7.4 (Joomla 3.9.x / 4.x)
 * @author      Stergios Zgouletas <info@web-expert.gr>
 * @link        http://www.web-expert.gr
 * @copyright   Copyright (C) 2018 Web-Expert.gr All Rights Reserved
 * @license     http://www.gnu.org/licenses/gpl-2.0.html GNU/GPL
 */
 
defined( '_JEXEC' ) or die( 'Restricted access' );

use Joomla\CMS\Factory; 
use Joomla\CMS\Plugin\CMSPlugin as JPlugin;
use Joomla\String\StringHelper as JString;
use Joomla\CMS\Uri\Uri;

class plgSystemServercache extends JPlugin
{
	protected $app; 
	
	private $debug=array();
	private $cacheSwOff=0;
	protected $httpPush=array();
	
	public function __construct(&$subject, $config)
	{
		parent::__construct($subject, $config);
		
		if(!$this->app)
		{
			$this->app = JFactory::getApplication();
		}
	}

	private function JCacheSwitch($mode)
	{
		$this->app->set('caching',$mode);
	}	
	
	private function isDebug()
	{
		if($this->app->input->get('format','html')!='html') return 0;
		
		static $debugStatus=null;
		if(is_null($debugStatus))
		{
			$clientIP=$this->getUserIP();
			$debugStatus=($this->params->get('debug',0)==1 || ($this->params->get('debug',0)==2 && $this->params->get('ip','')==$clientIP));
		}
		return $debugStatus;
	}
	
	function onAfterRoute()
	{
		if($this->params->get('event','')=='onAfterRoute' && !$this->app->isClient('cli'))
		{
			$this->serverCache();
		}
	}
	
	function onAfterDispatch()
	{
		if($this->params->get('event','')=='onAfterDispatch' && !$this->app->isClient('cli'))
		{
			$this->serverCache();
			if($this->cacheSwOff) $this->JCacheSwitch($this->cacheSwOff);
		}
	}
	
	function onBeforeRender()
	{
		if($this->params->get('event','')=='onBeforeRender' && !$this->app->isClient('cli'))
		{
			$this->serverCache();
		}
	}
	
	function onAfterRender()
	{
		if($this->app->input->getCmd('format','html')!='html' || $this->app->isClient('cli')) return;
		
		$buffer=version_compare(JVERSION,'3.6','ge')?$this->app->getBody():JResponse::getBody();		
		
		if($this->params->get('http2_push',0) && $this->app->isSSLConnection())
		{
			$siteUrl=rtrim(Uri::root(),'/');
			$host=parse_url($siteUrl,PHP_URL_HOST);
			$debugURLs=array();
			
			if(preg_match_all('/src="(.*)"/isU',$buffer,$matches))
			{
				foreach($matches[1] as $idx=>$url)
				{
					if(empty($url) || JString::strpos($url,'=feed')!==false || JString::strpos($url,'googleapis')!==false) continue;
					
					if(preg_match("/^.*\.(jpg|jpeg|png|gif|ico|bmp|svg|webp)$/i",$url,$extMatch))
					{
						$type='image';
					}
					elseif(preg_match("/^.*\.(js)$/i",$url,$extMatch))
					{
						$type='script';
					}
					else
					{
						continue;
					}
					
					$urlNew=$url;
					if(JString::substr($url,0,2)=='//') $urlNew='https://'.JString::substr($url,2);
					
					$debugURLs['media'][]=$urlNew;

					if(JString::strpos($urlNew,'http')!==false && $host!=parse_url($urlNew,PHP_URL_HOST)) continue;

					$urlNew=str_replace($siteUrl,'',$urlNew);
					if(JString::substr($urlNew,0,1)!=='/') $urlNew='/'.$urlNew;

					$debugURLs['media_pushed'][]=$urlNew;

					//$buffer=str_replace($url,$urlNew,$buffer);
					$this->addToPreload($url,$type);
				}
			}
			
			if(preg_match_all('/<link(.*)href="(.*)"/isU',$buffer,$matches))
			{
				foreach($matches[2] as $idx=>$url)
				{
					if(empty($url) || JString::strpos($url,'.css')===false) continue;
					$urlNew=$url;
					if(JString::substr($url,0,2)=='//') $urlNew='https://'.JString::substr($url,2);
		
					$debugURLs['link'][]=$urlNew;
					if(JString::strpos($urlNew,'http')!==false && parse_url($siteUrl,PHP_URL_HOST)!=parse_url($urlNew,PHP_URL_HOST))
					{
						continue;
					}
					
					$urlNew=str_replace($siteUrl,'',$urlNew);
					if(JString::substr($urlNew,0,1)!=='/') $urlNew='/'.$urlNew;
					
					$debugURLs['link_pushed'][]=$urlNew;
					//$buffer=str_replace($url,$urlNew,$buffer);	
					$this->addToPreload($url,'style');
				}
			}
			
			if(count($this->httpPush))
			{
				$this->setHeader('Link',implode(', ',$this->httpPush),true);
				$this->httpPush=array();
				$this->sendHeaders();
			}
		}
		
		if($this->isDebug())
		{
			$debug='<pre>Disallow Caching:'.var_export(in_array(true,$this->debug['rules'],true),true).'|'.$this->debug['cachingMode'].' => '.print_r($this->debug,true).'</pre>';
			$buffer=str_replace('</body>',$debug.'</body>',$buffer);		
		}
		version_compare(JVERSION,'3.6','ge')?$this->app->setBody($buffer):JResponse::setBody($buffer);
		$this->debug=array();
	}
	
	function serverCache($format='auto')
	{
		$session = Factory::getSession();
		
		$off=(int)$session->get('servercacheoff',0) || defined('PLG_SERVER_CACHE_OFF'); //For developers		
		$this->debug['rules']=array();
		
		//disable caching on Adminstrator Page or on POST/PUT/HEAD/DELETE method
		if($this->app->isClient('administrator') || $this->app->input->getMethod()!='GET' || $this->params->get('devmode',0) || $off)
		{
			$this->addHeadersToPage('deny',$format);
			return;
		}
		
		$menu = $this->app->getMenu();
		$activeMenu=$menu->getActive();
		$Itemid=$this->app->input->get('Itemid',isset($activeMenu->id)?$activeMenu->id:0);
		
		$disableCache=array();
		$disableCache['access.level']=0;

		$defaultComponent=null;
		$defaultView=null;
		if($Itemid>0)
		{
			$cache = Factory::getCache();
			$cache->setCaching(1);
			if(version_compare(JVERSION,'4.0','ge'))
				$item = $cache->get(array($menu,'getItem'),$Itemid);
			else
				$item = $cache->call(array($menu,'getItem'),$Itemid);
			
			if(isset($item->query['option'])) $defaultComponent=$item->query['option'];
			if(isset($item->query['view'])) $defaultView=$item->query['view'];
			$disableAccess=$this->params->get('components',array(0));
			if(!in_array(0,$disableAccess) && in_array($item->access,$disableAccess)) $disableCache['access.level']=1;
		}
			
		$option=$this->app->input->get('option',$defaultComponent);
		$view=$this->app->input->get('view',$defaultView);
				
		//User Custom Checks
		$disableURLs=$this->explodeUrls('urls');
		$disableURLs[]='cart';
		$disableURLs[]='checkout';
		$disableURLs[]='register';
		$disableURLs[]='login';
		$disableURLs[]='orders';
		
		$disableURLs=array_unique(array_filter($disableURLs));
		
		if(!empty($disableURLs))
		{
			$uri = Uri::getInstance(); 
			$currentURL=$uri->toString();
			$this->debug['currentURL']=$currentURL;
			
			foreach($disableURLs as $url)
			{
				if(!empty($url) && JString::strpos($currentURL,$url)!==false)
				{
					$disableCache['url.'.$url]=1;
					break;
				}
			}
		}
		
		$components=$this->params->get('components',array("com_users"));
		$disableCompos=array_filter($components);
		
		if(!empty($disableCompos))
		{
			$disableCache['component.'.$option]=in_array($option,$disableCompos)?1:0;
		}
		
		$disableViews=JString::trim($this->params->get('views',"user\nlogin\nreset\nremind\nregistration"));
		if(!empty($disableViews))
		{
			$disableCache['view.'.$view]=in_array($view,explode("\n",$disableViews))?1:0;
		}
		
		$byPass=$this->params->get('cachebypass','jnocache');
		if(!empty($byPass)) $disableCache['bypass.keyword']=JString::strpos($currentURL,$byPass)!==false?1:0;
		
		//Common
		$disableCache['isLogged']=Factory::getUser()->guest?0:1;
		$disableCache['view_cart']=!empty($view) && JString::strpos($view,'cart')!==false?1:0; //ensure that cart is excluded, reason SEF
		$cachingMode=in_array(1,$disableCache)?'deny':'allow';
		
		//Check for products to Virtuemart cart
		if($this->params->get('checkcart',1) && $cachingMode=='allow')
		{
			//COM_VIRTUEMART
			if(JFile::exists(JPATH_ADMINISTRATOR.'/components/com_virtuemart/helpers/config.php'))
			{
				if(!defined('DS')) define('DS', DIRECTORY_SEPARATOR);
				if(!class_exists('VmConfig')) require_once(JPATH_ADMINISTRATOR.DS.'components'.DS.'com_virtuemart'.DS.'helpers'.DS.'config.php');
				VmConfig::loadConfig();
				if(!class_exists('VmModel')) require_once(JPATH_ADMINISTRATOR.DS.'components'.DS.'com_virtuemart'.DS.'helpers'.DS.'vmmodel.php');
				if(!class_exists('VirtueMartCart')) require_once(JPATH_ROOT.DS.'components'.DS.'com_virtuemart'.DS.'helpers'.DS.'cart.php');
				$cart=VirtueMartCart::getCart();
				$cart->prepareCartData();
				if($cart->totalProduct)
				{
					$disableCache['vm_cart_has_products']=1; //just for debug
					$cachingMode='deny';
				}
			}
			
			//COM_J2STORE
			if($cachingMode=='allow' && JFile::exists(JPATH_SITE.'/modules/mod_j2store_cart/helper.php'))
			{
				if(!defined('F0F_INCLUDED')) include_once JPATH_LIBRARIES . '/f0f/include.php';
				if(!class_exists('modJ2StoreCartHelper')) require_once(JPATH_SITE.'/modules/mod_j2store_cart/helper.php');
				$list = modJ2StoreCartHelper::getItems();
				if($list['product_count'])
				{
					$disableCache['j2store_cart_has_products']=1; //just for debug
					$cachingMode='deny';
				}	
			}
			
			//COM_ESHOP
			if($cachingMode=='allow' && JFile::exists(JPATH_SITE.'/administrator/components/com_eshop/libraries/defines.php'))
			{
				require_once JPATH_ROOT.'/administrator/components/com_eshop/libraries/defines.php';
				require_once JPATH_ROOT.'/administrator/components/com_eshop/libraries/inflector.php';
				require_once JPATH_ROOT.'/administrator/components/com_eshop/libraries/autoload.php';
				$cart = new EshopCart();
				$countProducts = $cart->countProducts();
				if($countProducts)
				{
					$disableCache['eshop_cart_has_products']=1; //just for debug
					$cachingMode='deny';
				}
			}
			
			//COM_REDSHOP
			if($cachingMode=='allow' && JFile::exists(JPATH_SITE.'/administrator/components/com_redshop/redshop.php'))
			{
				if (version_compare(JVERSION, '4.0', '<')) JLoader::import('redshop.library');
				
				\RedshopHelperUtility::databaseToCart();
				$totalQuantity = \Redshop\Cart\Helper::getTotalQuantity();
				if($totalQuantity)
				{
					$disableCache['redeshop_cart_has_products']=1; //just for debug
					$cachingMode='deny';
				}
			}			
		}
		
		$this->debug['rules']=$disableCache;
		$this->debug['cachingMode']=$cachingMode;
		//False: Dont allow to Cache
		//True: Allow Caching
		$this->addHeadersToPage($cachingMode,$format);
	}
	
	protected function addHeadersToPage($mode,$format='auto')
	{			
		$docType=Factory::getDocument()->getType();
		$format = empty($format) || $format=='auto'?strtolower($this->app->input->getWord('format',$docType)):$format;
		$this->debug['raw-format']=$format;
		
		if($format=='rss' || $format=='atom') $format='xml';
		$format=in_array($format,array('xml','json','html'))?$format:'html'; //anything else handle as HTML
		$this->debug['format']=$format;
		
		$langTag = Factory::getLanguage()->getTag();
		$menu = $this->app->getMenu();
		$activeMenu=$menu->getActive();
		$isFrontPage=$activeMenu == $menu->getDefault($langTag); //multilanguage detection	
		$cacheTime=$isFrontPage?$this->params->get('cachingtimehome',60):$this->params->get('cachingtime'.$format,300); //seconds
		
		
		$this->debug['isAdmin']=$this->app->isClient('administrator')?1:0;
		$this->debug['GET-Method']= $this->app->input->getMethod()=='GET'?1:0;
		$this->debug['CachingTime']=$cacheTime.' seconds';
		$this->debug['isFrontPage']=$isFrontPage?'Yes':'No'; 
		$this->debug['developmentMode']=$this->params->get('devmode',0);
		
		//Disable Joomla cache
		if($mode!='allow')
		{
			$this->cacheSwOff=(int)$this->app->get('caching');
			$this->JCacheSwitch(0);
			$this->debug['JoomlaCacheBefore']=$this->cacheSwOff;
		}
		$this->debug['JoomlaCacheNow']=$this->app->get('caching');
		
		$session = Factory::getSession();
		$this->debug['servercacheoff-Session']=(int)$session->get('servercacheoff',0) || defined('PLG_SERVER_CACHE_OFF');
	
		version_compare(JVERSION,'3.6','ge')? $this->app->allowCache(true) : JResponse::allowCache(true);
		
		//CACHINHG HEADERS
		if($mode=='allow' && $cacheTime)
		{
			$this->setHeader('Cache-Control','public, max-age='.$cacheTime.', pre-check='.$cacheTime.', post-check='.$cacheTime.', stale-while-revalidate='.($cacheTime*2).', stale-if-error='.($cacheTime*3));
			$this->setHeader('Pragma', 'public');
			$this->setHeader('Expires', gmdate('D, d M Y H:i:s', time()+$cacheTime).' GMT');
			$this->setHeader('X-Logged-In','False');
		}
		else
		{
			$this->setHeader('Cache-Control', 'private, max-age=0, no-cache, no-store');
			$this->setHeader('Pragma', 'no-cache');
			$this->setHeader('Expires', 'Mon, 01 Jan 2019 00:00:00 GMT');
			$this->setHeader('X-Logged-In','True');
		}
		
		//CORS Headers 
		$corsmode=(int)$this->params->get('corsmode',0);
		if($corsmode && !$this->app->isClient('administrator'))
		{
			$originVal='*';
			if($corsmode==2)
			{
				$rqOrigin=$this->getBestOrigin();
				$this->debug['Origin-Requested']=$rqOrigin;
				
				$origins=$this->explodeUrls('acndomains');
				$url=parse_url(Uri::root()); //may includes the path
				$origins[]=$url['scheme'].'://'.$url['host'];
				$originVal= implode( ', ', $origins );
				/*if(!empty($origins) && in_array($rqOrigin,$origins))
				{
					$originVal=$rqOrigin;
				}
				else
				{
					$url=parse_url(Uri::root()); //may includes the path
					$originVal=$url['scheme'].'://'.$url['host'];
				}*/
				$this->setHeader('Vary','Origin');
			}
			$this->debug['cors']=$originVal;
			$this->setHeader('Access-Control-Allow-Origin',$originVal);
		}
		
		$cspMode=(int)$this->params->get('cspmode',0);
		if($cspMode && !$this->app->isClient('administrator'))
		{
			if($cspMode==2)
			{
				//CSP Headers 
				$maindomain='*'.str_replace('www.','.',$_SERVER['SERVER_NAME']); //allow subdomains
				$popularCDNs=array(
					$maindomain,
					'*.twitter.com',
					'*.facebook.net',
					'*.facebook.com',
					'*.dailymotion.com',
					'*.youtube.com',
					'*.bootstrapcdn.com',
					'*.cloudflare.com',
					'*.gstatic.com',
					'*.google.com',
					'*.googleapis.com',
					'*.google-analytics.com',
					'*.googletagmanager.com',
					'code.jquery.com',
					'ajax.aspnetcdn.com',
					'www.w3.org',
					'use.fontawesome.com',
				);
				
				//CSP only works in modern browsers Chrome 25+, Firefox 23+, Safari 7+
				$headerCSP=array();
				$types=array('connect-src','connect-src','default-src','frame-ancestors','frame-src','img-src','media-src','object-src','script-src','style-src');
				foreach($types as $type)
				{
					$data=$this->explodeUrls($type,$popularCDNs);
					$data=in_array('*',$data)?"'self' 'unsafe-inline' *":implode(' ',$data);
					$data=str_replace(array('self','unsafe-inline','none','strict-dynamic',"''","\n"),array("'self'","'unsafe-inline'","'none'","'strict-dynamic'","'",''),$data);
					$headerCSP[$type]=$type.' '.$data.';'."\n";
				}
				$this->debug['csp']=$headerCSP;
				$headerCSP=implode('',$headerCSP);
			}
			else
			{
				//allow  All
				$headerCSP = "default-src *  data: blob: 'unsafe-inline'; script-src * 'unsafe-inline'; connect-src * 'unsafe-inline'; img-src * data: blob: 'unsafe-inline'; frame-src *; style-src * data: blob: 'unsafe-inline';font-src * data: blob: 'unsafe-inline';";
				$this->debug['csp']=$headerCSP;
			}
			
			$this->setHeader("Content-Security-Policy",$headerCSP);
		}
		
		/*$encodingMode=(int)$this->params->get('encodingmode',0);
		if($encodingMode)
		{
			$supportsGzip = strpos($_SERVER['HTTP_ACCEPT_ENCODING'], 'gzip' ) !== false;
			$supportsDeflate = strpos($_SERVER['HTTP_ACCEPT_ENCODING'], 'deflate') !== false;
			
			if($supportsDeflate && $encodingMode==1)
				$this->setHeader('Content-Encoding','deflate',true);
			else if($supportsGzip && $encodingMode==2) 
				$this->setHeader('Content-Encoding','gzip',true);
			//else if($supportsGzip && $encodingMode==3) 
				//$this->setHeader('Content-Encoding','gzip, deflate');
		}*/
		
		$this->setHeader('WE-SERVERCACHE','Powered by Web-expert.gr');
		$this->sendHeaders();
	}
	
	protected function setHeader($name,$value,$replace=true)
	{
		if(version_compare(JVERSION,'3.6','ge'))
		{
			$this->app->setHeader($name,$value,$replace);
		}
		else
		{
			JResponse::setHeader($name,$value,$replace);
		}
		if($this->isDebug()) $this->debug['headers'][$name]=htmlspecialchars($value);
	}
	
	protected function sendHeaders()
	{
		if(version_compare(JVERSION,'3.6','ge'))
		{
			$this->app->sendHeaders();
		}
		else
		{
			JResponse::sendHeaders();
		}
		$this->debug['headerSent']=1;
	}
	
	protected function getBestOrigin()
	{
		$origin = $this->getUserIP();
		if (array_key_exists('HTTP_ORIGIN', $_SERVER))
		{
			$origin = $_SERVER['HTTP_ORIGIN'];
		}
		else if (array_key_exists('HTTP_REFERER', $_SERVER))
		{
			$origin = $_SERVER['HTTP_REFERER'];
		}
		
		return $origin;
	}
	
	protected function explodeUrls($paramName,$append=array())
	{
		$URLs=JString::trim(JString::trim($this->params->get($paramName,''),"\n"));
		$URLs=explode("\n",$URLs);
		return array_filter(array_unique(array_merge($URLs,$append)));
	}
	
	protected function addToPreload($url, $type)
	{
		if(isset($this->httpPush[$url])) return false;
		
		$preload = '<'.$url.'>; rel=preload; as='.$type;

		if($type == 'font')
		{
			$preload .= '; crossorigin';
			if(preg_match("#\.\K\w++(?=$|[\#?])#", $url, $m))
			{

				switch ($m[0])
				{
					case 'woff2':
						$preload .= '; type="font/woff2"';
						break;
					case 'woff':
						$preload .= '; type="font/woff"';
						break;
					case 'eot':
						$preload .= '; type="application/vnd.ms-fontobject"';
						break;
					case 'otf':
						$preload .= '; type="font/otf"';
						break;
					case 'ttf':
						$preload .= '; type="font/ttf"';
						break;
					case 'svg':
						$preload .= '; type="image/svg+xml"';
						break;
					default:
						break;
				}
			}
		}
		$this->httpPush[$url]=$preload;
		//$this->debug['http2'][]=htmlspecialchars($preload);
	}
	
	function getUserIP()
	{
		$ipaddress = '';
		if(isset($_SERVER['HTTP_CF_CONNECTING_IP']))
			$ipaddress = $_SERVER['HTTP_CF_CONNECTING_IP'];	//cloudflare real IP
		else if (isset($_SERVER['HTTP_CLIENT_IP']))
			$ipaddress = $_SERVER['HTTP_CLIENT_IP'];
		else if(isset($_SERVER['HTTP_X_FORWARDED_FOR']))
			$ipaddress = $_SERVER['HTTP_X_FORWARDED_FOR'];
		else if(isset($_SERVER['HTTP_X_FORWARDED']))
			$ipaddress = $_SERVER['HTTP_X_FORWARDED'];
		else if(isset($_SERVER['HTTP_X_CLUSTER_CLIENT_IP']))
			$ipaddress = $_SERVER['HTTP_X_CLUSTER_CLIENT_IP'];
		else if(isset($_SERVER['HTTP_FORWARDED_FOR']))
			$ipaddress = $_SERVER['HTTP_FORWARDED_FOR'];
		else if(isset($_SERVER['HTTP_FORWARDED']))
			$ipaddress = $_SERVER['HTTP_FORWARDED'];
		else if(isset($_SERVER['REMOTE_ADDR']))
			$ipaddress = $_SERVER['REMOTE_ADDR'];
		else
			$ipaddress = 'UNKNOWN';
		return $ipaddress;
	}
	
}